import os
import os.path
import stat


class PathSet(object):

    """Represents an ordered set of unique normalized file paths.
        The paths may be the subject to user (~) and variable ($) expansion.
        This class fulfills the sequence contract (i.e. len(), x[], iterating etc.).
    """

    def __init__(self, *args, **kwds):
        """Construct a set. Any number of either string-like objects or the sequences of
            thereof is accepted.
            The following keywords are recognized:
                userExpand    -- (default True)
                varExpand    -- (default True)

        """
        self.paths = []
        try:
            self.varExpand = kwds["varExpand"]
        except KeyError:
            self.varExpand = True
        try:
            self.userExpand = kwds["userExpand"]
        except KeyError:
            self.userExpand = True
        for x in args:
            self.append(x)

    def __iter__(self):
        return self.paths.__iter__()

    def __len__(self):
        return len(self.paths)

    def __str__(self):
        paths = self.paths
        if len(paths) > 1:
            s = str(paths[0])
            for x in paths[1:]:
                s += ":" + str(x)

        elif len(paths) > 0:
            s = str(paths[0])
        else:
            s = ""
        return s

    def __contains__(self, path):
        return self.adopt(path) in self.paths

    def __getitem__(self, index):
        return self.paths[index]

    def __add__(self, other):
        x = self.__class__()
        x.append(self)
        x.append(other)
        return x

    def __iadd__(self, other):
        self.append(other)
        return self

    def adopt(self, path):
        """Validates the specified path and converts it to a form suitable for inclusion
            into the set. Must either return the path or raise ValueError if an inappropriate
            path was given.
        """
        if self.userExpand and path.startswith("~"):
            path = os.path.expanduser(path)
        if self.varExpand and "$" in path:
            path = os.path.expandvars(path)
        path = os.path.normpath(path)
        if path in self.paths:
            raise ValueError
        else:
            return path

    def append(self, arg):
        if isinstance(arg, str):
            try:
                self.paths.append(self.adopt(arg))
            except ValueError:
                pass

        else:
            for x in arg:
                self.append(str(x))  # WARNING : a recursive call, stay alert


class ValidPathSet(PathSet):

    """Represents the PathSet of the valid paths. The symbolic links are respected
        and have the preference over the real directories they point to.
    """

    def adopt(self, path):
        path = super(ValidPathSet, self).adopt(path)
        try:
            st = os.stat(path)
            for xst in self.stats:
                if os.path.samestat(st, xst):
                    # FIXME : inefficiency
                    if os.path.islink(path) and not os.path.islink(
                            self.stats[xst]):
                        # In case of real path followed by the symlink to it
                        # the latter has the preference
                        self.paths[self.paths.index(self.stats[xst])] = path
                        del self.stats[xst]
                        self.stats[st] = path
                    # In all other cases the path is discarded
                    raise ValueError  # To prevent the appending of the path

            self.stats[st] = path
            return path
        except OSError:
            raise ValueError

    def __init__(self, *args, **kwds):
        self.stats = {}
        super(ValidPathSet, self).__init__(*args, **kwds)



import stat
import sys
import os.path
import MenuMaker
import Prophet

from Keywords import Keyword as Kw, Set as KwS

import optparse
from optparse import OptionParser as _OP

from MenuMaker import msg, warn, fatal, fronts, terms

# Output the generated menu to stdout instead of the menu file
writeToStdout = False
forceWrite = False  # Overwrite existing files
noDesktop = False  # Do not scan for .desktop files
noLegacy = False  # Do not scan for the legacy apps
noDebian = False  # Exclude Debian apps database

vStr = "be verbose"


def vOpt(option, opt, value, parser):
    v = MenuMaker.verbose
    if v < 3:
        v += 1
        MenuMaker.verbose = Prophet.verbose = v


fStr = "overwrite existing files"


def fOpt(option, opt, value, parser):
    global forceWrite
    forceWrite = True

cStr = "dump the menu to stdout instead of file"


def cOpt(option, opt, value, parser):
    global writeToStdout
    writeToStdout = True

iStr = "produce the output suitable for inclusion into the custom menu; MUST be used in conjunction with -c"


def iOpt(option, opt, value, parser):
    MenuMaker.writeFullMenu = False

pStr = "always retain the path to executable (by default, the path is omitted whenever possible)"


def pOpt(option, opt, value, parser):
    Prophet.omitExecPath = False

xdStr = "skip .desktop database scanning"


def xdOpt(option, opt, value, parser):
    global noDesktop
    noDesktop = True

xlStr = "skip legacy knowledge base scanning"


def xlOpt(option, opt, value, parser):
    global noLegacy
    noLegacy = True

xeStr = "skip Debian database scanning"


def xeOpt(option, opt, value, parser):
    global noDebian
    noDebian = True

sKwS = KwS(*Prophet.skip.keys())
sStr = "skip specific categories (comma-separated list)"


def sOpt(option, opt, value, parser):
    for x in value.split(","):
        if not len(x):
            continue
        x = Kw(x)
        try:
            Prophet.skip[x]
        except KeyError:
            raise optparse.OptionValueError(
                "wrong argument %s for option %s" % (x, opt))
        Prophet.skip[x] = True


tStr = "terminal emulator to use for console applications"


def tOpt(option, opt, value, parser):
    global term
    for t, n in terms:
        if value in n:
            term = t
            return

    raise optparse.OptionValueError(
        "wrong argument %s for option %s" % (value, opt))


class OP(_OP):

    def print_help(self):
        sys.stdout.write(
            "This is %s %s, a 100%% Python heuristics-driven menu generator\n" %
            (MenuMaker.pkgName, MenuMaker.pkgVer))
        sys.stdout.write(
            "%s %s %s <%s>\n\n" %
            (MenuMaker.pkgHome,
             MenuMaker.copyright,
             MenuMaker.author,
             MenuMaker.email))
        _OP.print_help(self)
        sys.stdout.write("\nfrontends (case insensitive):\n")
        for v in fronts.values():
            vars = list(v)
            s = str(vars[0])
            for x in vars[1:]:
                s += " | " + str(x)
            sys.stdout.write("  %s\n" % s)
        sys.stdout.write(
            "\nterminal emulators for -t (case insensitive), in order of decreasing preference:\n")
        for t, n in terms:
            vars = list(n)
            s = str(vars[0])
            for x in vars[1:]:
                s += " | " + str(x)
            sys.stdout.write("  %s\n" % s)
        sys.stdout.write("\nskip categories for -s (case insensitive):\n")
        for k in sKwS:
            sys.stdout.write("  %s\n" % k)


op = OP(usage="%prog [options] frontend", version="%s %s" %
        (MenuMaker.pkgName, MenuMaker.pkgVer))
op.add_option("-v", "--verbose", action="callback", callback=vOpt, help=vStr)
op.add_option("-f", "--force", action="callback", callback=fOpt, help=fStr)
op.add_option("-c", "--stdout", action="callback", callback=cOpt, help=cStr)
op.add_option("-i", action="callback", callback=iOpt, help=iStr)
op.add_option("-t", metavar="terminal", nargs=1, type="string",
              action="callback", callback=tOpt, help=tStr)
op.add_option(
    "-p", "--retain-path", action="callback", callback=pOpt, help=pStr)
op.add_option("--no-desktop", action="callback", callback=xdOpt, help=xdStr)
op.add_option("--no-legacy", action="callback", callback=xlOpt, help=xlStr)
op.add_option("--no-debian", action="callback", callback=xeOpt, help=xeStr)
op.add_option("-s", "--skip", metavar="list", nargs=1,
              type="string", action="callback", callback=sOpt, help=sStr)

opts, args = op.parse_args()

if not len(args):
    fatal("no frontend specified")

found = False
for k, v in fronts.items():
    if args[0] in v:
        found = True
        name = "MenuMaker.%s" % k
        frontName = k
        __import__(name)
        front = sys.modules[name]
        MenuMaker.setFrontend(front)
        break


if not found:
    fatal("unknown frontend: %s" % args[0])

if len(args) > 1:
    warn("multiple frontends specified; the first one (%s) will be used" %
         frontName)

try:
    menuFile = front.menuFile

except AttributeError:
    if not writeToStdout:
        fatal(
            "%s doesn't support writing to a file; use -c option instead" %
            frontName)


if not MenuMaker.writeFullMenu and not writeToStdout:
    fatal(
        "-i MUST be used in conjunction with -c; this is done to prevent accidental overwriting of the custom menu")

# Postpone the setup to the time when it's actually needed
msg("* scanning")

if noDesktop:
    msg("  skipping desktop")
    desktop = []
else:
    import Prophet.Desktop
    desktop = Prophet.Desktop.scan()

if noLegacy:
    msg("  skipping legacy")
    legacy = []
else:
    import Prophet.Legacy
    Prophet.Legacy.setup()
    legacy = Prophet.Legacy.scan()

if noDebian:
    msg("  skipping Debian")
    debian = []
else:
    import Prophet.Debian
    debian = Prophet.Debian.scan()

merged = Prophet.merge(legacy + desktop + debian)

msg("* generating")

if not Prophet.skip[Prophet.App.Console]:
    try:
        MenuMaker.terminal = term()
    except Prophet.NotSet:
        fatal("specified terminal emulator couldn't be found; try another one")
    except NameError:
        warn("  no terminal emulator specified; will use the default")
        found = False
        for tcls, tname in terms:
            try:
                MenuMaker.terminal = tcls()
                found = True
                break
            except Prophet.NotSet:
                pass

        if not found:
            fatal("no suitable terminal emulator found")

    msg("  using %s as terminal emulator" % MenuMaker.terminal.name)

menu = MenuMaker.Root()

if writeToStdout:
    msg("* writing")
    file = sys.stdout

else:
    msg("* writing to %s" % menuFile)
    file = os.path.expanduser(menuFile)
    dir = os.path.dirname(file)
    if not os.path.isdir(dir):
        msg("  creating missing directory %s" % dir)
        os.makedirs(dir)
    if not forceWrite and os.path.isfile(file):
        fatal(
            "refuse to overwrite existing file %s; either delete it or use -f option" %
            file)
    file = open(file, "wt")

menu.distribute(merged)

menu.arrange()

for x in menu.emit(0):
    file.write(x)
    file.write("\n")

msg("* done")



import re
import os.path
import glob
import fnmatch
import codecs
import Prophet

from Paths import ValidPathSet as PS
from Keywords import Keyword as Kw, Set as KwS
from Prophet import NotSet
from Prophet.Categories import *
from Prophet import msg, verbose


class DebianSet(PS):

    def adopt(self, path):
        path = super(DebianSet, self).adopt(path)
        if len(glob.glob(os.path.join(path, "*"))):
            return path  # Directory should not be empty
        raise ValueError


dirs = DebianSet(("/etc/menu", "/usr/lib/menu",
                  "/usr/share/menu", "/usr/share/menu/default", "~/.menu"))

# If dirs isn't empty then assume we're running Debian and therefore it's reasonable
# to use local menu entries instead of cached ones
if not len(dirs):
    dirs = DebianSet((os.path.join(__path__[0], "menu"),))
    runDebian = False

else:
    runDebian = True


def scan():
    """Scan through all containers and return a list of found valid entries"""
    result = []
    msg("  debian...", newline=False)
    for d in dirs:
        for f in glob.glob(os.path.join(d, "*")):
            if verbose > 1:
                msg("\nparsing " + f)
            try:
                entries = _parse(f)
            except NotSet as e:
                if verbose > 1:
                    msg("REJECTED : " + str(e))
            else:
                for e in entries:
                    try:
                        result.append(App(e))
                    except Prophet.NotSet as e:
                        if verbose > 1:
                            msg("REJECTED : " + str(e))

    msg(" %d apps found" % len(result))
    return result


def _parse(debmenu):
    try:
        # Latest Py3 versions in 64-bit mode sometimes have problems decoding Debian menus
        # complaining about malformed utf-8 codes, therefore one has to enforce the 8-bit encoding.
        # encoding= is not supported by Py2, so use codecs.open() as the least
        # common denominator.
        file = codecs.open(debmenu, "r", "latin1")  # Py 2.4+
    except IOError:
        raise NotSet("couldn't read the file " + debmenu)
    entry = ""
    entries = []
    next = False
    try:
        for x in file:
            x = x.strip()
            if next:
                if not len(x):
                    next = False
                    continue
                if not x.startswith("#"):
                    if x.endswith("\\"):
                        entry += " " + x.rstrip("\\")
                        next = True
                    else:
                        entry += " " + x
                        next = False

            else:
                if not len(x):
                    continue
                if not x.startswith("#") and x.startswith("?"):
                    if len(entry):
                        entries += _parseEntry(entry)
                    if x.endswith("\\"):
                        entry = x.rstrip("\\")
                        next = True
                    else:
                        entry = x
                        next = False

    except IOError:
        raise NotSet("i/o error while reading the file " + debmenu)
    except StopIteration:
        pass
    if len(entry):
        entries += _parseEntry(entry)
    return entries

_package = re.compile('^\?\s*?package\s*?\((.*?)\)\s*?\:')
_key = re.compile('(\w+)\s*=\s*(".*?"|\S+)')


def _parseEntry(s):
    entry = {}
    x = _package.match(s)
    if x:
        entry["package"] = x.group(1)
        for x in re.findall(_key, s):
            entry[x[0]] = x[1].strip('"')

    return [entry]


class App(Prophet.App):

    pref = 10

    def __new__(cls, entry):
        self = object.__new__(cls)
        self.__setup__(entry)
        return self

    def __setup__(self, entry):
        self.entry = entry
        super(App, self).__setup__()
        del self.entry
        # TODO : filter out GNOME/KDE apps since they should be picked up by the .desktop scanner anyways
        # To accomplish this scan title/longtitle/description/whatever for the
        # clues (gnome-*, "Calculator for GNOME", etc.)

    def setName(self):
        try:
            self.name = self.entry["title"]
        except KeyError:
            super(App, self).setName()

    def setComment(self):
        try:
            self.comment = self.entry["longtitle"]
        except KeyError:
            try:
                self.comment = self.entry["description"]
            except KeyError:
                super(App, self).setComment()

    def setKeywords(self):
        try:
            sect = self.entry["section"]
        except KeyError:
            sect = ""
        try:
            hint = self.entry["hints"]
        except KeyError:
            hint = ""
        kws = _deb2kws(sect, hint)
        if not len(kws):
            raise Prophet.NotSet("no keywords guessed")
        self.keywords = kws

    def setTerminal(self):
        try:
            needs = Kw(self.entry["needs"])
            if needs == Kw("X11"):
                self.terminal = False
            elif needs == Kw("text"):
                self.terminal = True
            else:
                raise Prophet.NotSet("unsupported needs entry (%s)" % needs)

        except KeyError:
            super(App, self).setTerminal()

    def setExename(self):
        try:
            cmd = self.entry["command"]
        except KeyError:
            raise Prophet.NotSet("no command entry found")
        self.testGoodness(cmd)
        scmd = cmd.split(" ", 1)
        if len(scmd) >= 2:
            cmd = scmd[0].strip()
            args = scmd[1].strip()
            for x in args.split():
                if fnmatch.fnmatch(x, "*/*/*") and not os.path.exists(x):
                    # An argument specified in the command line looks like a path. If it
                    # doesn't exist, the entire command is likely to be
                    # worthless, so throw it off
                    raise Prophet.NotSet(
                        "nonexistent path %s as command argument")

            self.exeargs = args
        if not runDebian:
            # If we're not running Debian specified command may be located
            # elsewhere therefore we have to scan PATH for it
            cmd = os.path.basename(cmd)
            # List of globs of unwanted commands
            if cmd in (
                    "sh",
                    "bash",
                    "csh",
                    "tcsh",
                    "zsh",
                    "ls",
                    "killall",
                    "echo",
                    "nice",
                    "xmessage",
                    "xlock"):
                # Debian seems to have lots of shell commands. While it would be nice
                # to try to analyze them, this isn't a priority since they're often
                # too Debian-specific.
                raise Prophet.NotSet("%s blacklisted" % cmd)

        self.exename = cmd


_secs = {
    Kw("Appearance"): (Settings, DesktopSettings),
    Kw("AI"): (Science,),
    Kw("Databases"): (Database, Office),
    Kw("Editors"): (TextEditor, Utility),
    Kw("Educational"): (Education,),
    Kw("Emulators"): (Emulator,),
    Kw("Games"): (Game,),
    Kw("Adventure"): (AdventureGame, Game),
    Kw("Arcade"): (ArcadeGame, Game),
    Kw("Board"): (BoardGame, Game),
    Kw("Card"): (CardGame, Game),
    Kw("Puzzles"): (LogicGame, Game),
    Kw("Simulation"): (Simulation, Game),
    Kw("Sports"): (SportsGame, Game),
    Kw("Strategy"): (StrategyGame, Game),
    Kw("Tetris-like"): (BlocksGame, Game),
    Kw("Graphics"): (Graphics,),
    Kw("Hamradio"): (HamRadio, Network),
    Kw("Math"): (Math, Science),
    Kw("Misc"): (Utility,),  # ???
    Kw("Net"): (Network,),
    Kw("Analog"): (Engineering, Network),
    Kw("Headlines"): (News, Network),
    Kw("Programming"): (Development,),
    Kw("Shells"): (Shell, ConsoleOnly),  # ???
    Kw("Sound"): (Audio, AudioVideo),
    Kw("System"): (System,),
    Kw("Admin"): (Settings, System),

    #    Kw("GNOME") : (GNOME, GTK),
    Kw("Technical"): (Engineering,),
    Kw("Text"): (TextEditor, Utility),
    Kw("Tools"): (Utility,),
    Kw("Viewers"): (Viewer, Graphics),
    Kw("XawTV"): (TV, Viewer, Video, AudioVideo),
    Kw("Toys"): (Amusement,),
    Kw("Help"): (Utility, Kw("X-Help")),
    Kw("Screen"): (Utility,),  # ???

    #    Kw("Lock") : (Utility, Kw("X-Lock")),
    Kw("Saver"): (Screensaver, Utility),
    Kw("System"): (System,),
    Kw("XShells"): (Shell,)
}

_hints = {
    Kw("3D"): (ActionGame, Game),  # ???
    Kw("Bitmap"): (RasterGraphics, x2DGraphics, Graphics),
    Kw("Boot"): (System, Utility),
    Kw("Bug reporting"): (Development,),
    Kw("Calculators"): (Calculator, Utility),
    Kw("Clocks"): (Clock, Utility),
    Kw("Config"): (Settings,),
    Kw("Debuggers"): (Debugger, Development),
    Kw("Documents"): (Office,),
    Kw("Doom"): (ActionGame, Game),
    Kw("Drawing"): (RasterGraphics, x2DGraphics, Graphics),
    Kw("Vector"): (VectorGraphics, Graphics),
    Kw("Equation"): (Math, Science),
    Kw("Editor"): (WordProcessor, Office),
    Kw("Formula"): (Presentation, Office),
    Kw("Fonts"): (DesktopSettings, Settings),
    Kw("Mail"): (Email, Office, Network),
    Kw("Calendar"): (Calendar, Office),
    Kw("Monitoring"): (Monitor, System),
    Kw("IRC"): (IRCClient, Network),
    Kw("ISDN"): (Telephony, Dialup, Network),  # ???
    Kw("Images"): (x2DGraphics, Graphics),
    Kw("Internet"): (Network,),
    Kw("HTML"): (Network,),
    Kw("Mahjongg"): (LogicGame, Game),
    Kw("Mines"): (LogicGame, Game),
    Kw("Mixers"): (Mixer, Audio),
    Kw("Real-time"): (StrategyGame, Game),
    Kw("Parallel"): (Math, Science, Network),
    Kw("Distributed"): (Math, Science, Network),
    Kw("PostScript"): (Presentation, Office),
    Kw("Presentation"): (Presentation, Office),
    Kw("SameGame"): (LogicGame, Game),
    Kw("Screenshot"): (Graphics, Utility),
    Kw("Setup"): (System,),
    Kw("Install"): (System,),
    Kw("Config"): (Settings,),
    Kw("Spreadsheets"): (Spreadsheet, Office),
    Kw("Terminal"): (TerminalEmulator),
    Kw("Translation"): (Languages, Education),
    Kw("Dictionary"): (Languages, Education),
    Kw("Time"): (Clock, Utility),
    Kw("Users"): (System, Settings),
    Kw("VNC"): (Network,),
    Kw("Video"): (Video, AudioVideo),
    Kw("Web Browsers"): (WebBrowser, Network),
    Kw("Word Processors"): (WordProcessor, Office)
}

_rejs = (
    Kw("Modules"), Kw("WindowManagers"), Kw(
        "WorkSpace"), Kw("Lock"), Kw("GNOME")
)


def _deb2kws(sec, hint):
    """Convert Debian style section/hints entries to an equivalent keyword set"""
    kws = []
    for x in sec.split("/"):
        x = Kw(x)
        if x in _rejs:
            raise Prophet.NotSet("section %s rejected" % x)
        try:
            kws += _secs[x]
        except KeyError:
            pass

    for x in hint.split(","):
        x = Kw(x)
        try:
            kws += _hints[x]
        except KeyError:
            pass

    return KwS(*kws)



from Prophet.Legacy import App as _App, ConsoleApp as _ConsoleApp, X11App as _X11App
from Keywords import Keyword as Kw, Set as KwS
from Prophet.Categories import *

X_HTML = Kw("X-HTML")


class asWedit(_App, _X11App):
    name = "asWedit"
    comment = "Advanced HTML editor"
    keywords = KwS(TextEditor, X_HTML)


class beaver(_App, _X11App):
    name = "Beaver"
    comment = "A programmer's editor"
    keywords = KwS(TextEditor, Development, GTK)


class bluefish(_App, _X11App):
    name = "BlueFish"
    comment = "Advanced HTML editor and web publishing application"
    keywords = KwS(TextEditor, Development, GNOME, X_HTML)


class cooledit(_App, _X11App):
    name = "CoolEdit"
    comment = "X11 text editor"
    keywords = KwS(TextEditor)


class cute(_App, _X11App):
    name = "Cute"
    comment = "User-friendly editor with syntax hilighting"
    keywords = KwS(TextEditor, Qt)


class ee(_App, _ConsoleApp):
    name = "Easy Editor"
    comment = "Standard FreeBSD text editor"
    keywords = KwS(TextEditor, Core)

    def valid(self, pfx, exe):
        import os.path
        if os.path.isfile(os.path.join(pfx, "share/misc/init.ee")):
            return super(ee, self).valid(pfx, exe)
        else:
            return False


class elvis(_App, _ConsoleApp):
    name = "Elvis"
    comment = "Vi compatible editor"
    keywords = KwS(TextEditor)


class emacs(_App, _ConsoleApp):  # FIXME : make this versioned
    name = "Emacs"
    comment = "A powerful text editing environment"
    keywords = KwS(TextEditor)

# TODO : jedit, jext


class jed(_App, _ConsoleApp):
    name = "Jed"
    comment = "Simple text editor"
    keywords = KwS(TextEditor)


class joe(_App, _ConsoleApp):
    name = "Joe"
    comment = "Joe's own editor"
    keywords = KwS(TextEditor)


class jove(_App, _ConsoleApp):
    name = "Jove"
    comment = "Emacs-like text editor"
    keywords = KwS(TextEditor)


class gvim(_App, _X11App):
    name = "GVim"
    comment = "X11 version of Vim, a Vi clone"
    keywords = KwS(TextEditor)


class gxedit(_App, _X11App):
    name = "GXedit"
    comment = "Simple X11 text editor"
    keywords = KwS(TextEditor, GTK)

# TODO : mp


class nano(_App, _ConsoleApp):
    name = "Nano"
    comment = "Simple text editor"
    keywords = KwS(TextEditor)


class nedit(_App, _X11App):
    name = "Nedit"
    comment = "Advanced X11 text editor"
    keywords = KwS(TextEditor, Motif)


class pico(_App, _ConsoleApp):
    name = "Pico"
    comment = "A Nano text editor clone"
    keywords = KwS(TextEditor)


class screem(_App, _X11App):
    name = "Screem"
    comment = "Advanced HTML editor"
    keywords = KwS(TextEditor, GTK, X_HTML)  # FIXME : GTK???


class vi(_App, _ConsoleApp):
    name = "Vi"
    comment = "Standard UNIX text editor"
    keywords = KwS(TextEditor, Core)


class vile(_App, _ConsoleApp):
    name = "Vile"
    comment = "Vi clone"
    keywords = KwS(TextEditor)


class vim(_App, _ConsoleApp):
    name = "Vim"
    comment = "Vi clone"
    keywords = KwS(TextEditor)


class we(_App, _ConsoleApp):  # FIXME : generic name, additional checks needed
    name = "We"
    comment = "Console version of Xwpe text editor"
    keywords = KwS(TextEditor, Development)


class xedit(_App, _X11App):
    name = "Xedit"
    comment = "Standard X11 text editor"
    keywords = KwS(TextEditor, Core)


class xemacs(_App, _X11App):  # FIXME : make this versioned
    name = "Xemacs"
    comment = "X11 version of Emacs"
    keywords = KwS(TextEditor)


class xjed(_App, _X11App):
    name = "Xjed"
    comment = "X11 version of Jed"
    keywords = KwS(TextEditor)


class xwe(_App, _X11App):
    name = "Xwe"
    comment = "X11 version of Xwpe text editor"
    keywords = KwS(TextEditor, Development)


class yudit(_App, _X11App):
    name = "Yudit"
    comment = "X11 unicode text editor"
    keywords = KwS(TextEditor)

# TODO : yzis



import MenuMaker

from MenuMaker import writeFullMenu

menuFile = "~/.deskmenurc"


def _map(x):
    for d, s in (("|", ":"),):
        x = x.replace(s, d)
    return x


class Sep(object):

    def emit(self, level):
        return ["divider=", ""]


class App(object):

    def emit(self, level):
        cmd = self.app.execmd
        if self.app.terminal:
            cmd = MenuMaker.terminal.runCmd(cmd)
        return ["menuitem=%s:%s" % (_map(self.app.name), cmd)]


class Menu(object):

    def emit(self, level):
        menu = ["submenu=%s" % _map(self.name)]
        for x in self:
            menu += x.emit(level + 1)
        menu += ["endmenu=", ""]
        return menu


class Root(object):
    name = "Deskmenu"

    def __init__(self, subs):
        if writeFullMenu:
            subs += [MenuMaker.Sep(), X("windowlist=Windows"),
                     X("workspaces=Workspaces")]
        super(Root, self).__init__(subs)

    def emit(self, level):
        if writeFullMenu:
            menu = ["keycode=Control+Escape", ""]
        else:
            menu = []
        for x in self:
            menu += x.emit(level)
        return menu


class X(MenuMaker.Entry):

    def __init__(self, x):
        super(X, self).__init__()
        self.align = MenuMaker.Entry.StickBottom
        self.x = x

    def emit(self, level):
        return [self.x]



import sys

if not (
        sys.version_info[0] *
        100 +
        sys.version_info[1] *
        10 +
        sys.version_info[2] >= 270):
    sys.exit("Python 2.7 or higher is required")

from Prophet import warn, fatal
from Keywords import Keyword as Kw, Set as KwS
import Prophet.Legacy.Shell as T

from Config import *

verbose = 0
omitEmptyMenu = True  # Whether to omit or to retain empty menu
# Whether to generate the entire menu or just a part to be included into
# the hand crafted one
writeFullMenu = True

fronts = {
    Kw("JWM"): KwS("JWM"),
    Kw("IceWM"): KwS("IceWM"),
    Kw("BlackBox"): KwS("BlackBox"),
    Kw("FluxBox"): KwS("FluxBox"),
    Kw("PekWM"): KwS("PekWM"),
    Kw("Deskmenu"): KwS("Deskmenu"),
    Kw("Xfce4"): KwS("Xfce", "Xfce4"),
    Kw("OpenBox3"): KwS("OpenBox", "OpenBox3"),
    Kw("WindowMaker"): KwS("WindowMaker", "WMaker")
}

# FIXME : the order should be frontend-specific
# For instance, Xfce may want to have Xfterm the preferable terminal
terms = [
    (T.xterm, KwS("Xterm")),
    (T._xfterm, KwS("Xfterm")),
    (T.xvt, KwS("Xvt")),
    (T.rxvt, KwS("Rxvt")),
    (T.aterm, KwS("Aterm")),
    (T.wterm, KwS("Wterm")),
    (T.Eterm, KwS("Eterm")),
    (T._gterm, KwS("GNOME-terminal", "Gterm")),
    (T._konsole, KwS("Konsole")),
    (T.sakura, KwS("Sakura")),
    (T.st, KwS("ST")),
    (T.alacritty, KwS("Alacritty")),
    (T.qterminal, KwS("QTerminal")),
]


def msg(s, newline=True):
    if verbose:
        sys.stderr.write(s)
        if newline:
            sys.stderr.write("\n")

    sys.stderr.flush()


def indent(level):
    """Return indentation string for specified level"""
    x = ""
    while(level > 2):
        x += "        "
        level -= 2
    for i in range(0, level):
        x += "    "
    return x


def setFrontend(module):
    """Specify the frontend to use"""
    for x in (Entry, Sep, App, Menu, Root):
        try:
            x.__bases__ = (getattr(module, x.__name__),) + x.__bases__
        except AttributeError:
            pass


class Entry(object):

    """Basic menu class"""
    # Alignment specifiers
    Floating = 0  # Element is subject to sorting
    # Element should be placed at the menu top retaining the original order
    StickTop = 1
    StickBottom = 2  # ...at the bottom...

    def __init__(self):
        self.align = Entry.Floating


class Sep(Entry):

    """A separator"""

    def __init__(self):
        self.align = Entry.StickBottom


class App(Entry):

    """Application entry"""

    def __init__(self, app):
        super(App, self).__init__()
        self.name = app.name
        self.app = app


class Menu(Entry, list):

    """Class which may contain other (sub)menus and single entries"""

    def __init__(self, subs):
        super(Menu, self).__init__()
        self[:] = subs

    def distribute(self, apps):
        """Distribute the application entries among the menu's submenus and then try to place the
            rest into itself. Return the list of the apps that didn't fit anywhere"""
        for x in self:
            if isinstance(x, Menu):
                apps = x.distribute(apps)

        rest = []
        for x in apps:
            if self.suitable(x):
                self.append(App(x))
            else:
                rest.append(x)

        return rest

    def empty(self):
        """Check whether the menu is empty, i.e. doesn't contain any meaningful entries
            (non-empty submenus, for ex.)"""
        for x in self:
            if isinstance(x, Menu):
                if not x.empty:
                    return False

            else:
                return False
            # TODO : separators do not count
        return True
    empty = property(empty)

    def emit(self, level):
        if omitEmptyMenu and self.empty:
            return []
        else:
            # Delegating it in hope that there is frontend-specific emit() down
            # the inheritance graph
            return super(Menu, self).emit(level)

    def suitable(self, app):
        """Simple suitability test -- check whether the app falls into any of specified categories"""
        return len(self.keywords & app.keywords) > 0

    def arrange(self):
        top = []
        bottom = []
        mfloat = []
        efloat = []
        for x in self:
            if isinstance(x, Menu):
                x.arrange()
            if x.align == Entry.Floating:
                if isinstance(x, Menu):
                    mfloat.append(x)
                else:
                    efloat.append(x)

            elif x.align == Entry.StickTop:
                top.append(x)
            elif x.align == Entry.StickBottom:
                bottom.append(x)

        mfloat.sort(key=Menu.extract)
        efloat.sort(key=Menu.extract)
        self[:] = top + mfloat + efloat + bottom
    #compare = staticmethod(lambda x, y : cmp(x.name.lower(), y.name.lower()))
    extract = staticmethod(lambda v: v.name.lower())

import os.path
import fnmatch

from Prophet.Categories import *
from Keywords import Set as KwS, Keyword as Kw


class Root(Menu):

    def __init__(self):
        subs = [
            DevApps, Editors, OfficeApps, ScienceApps, NetApps, GraphicApps,
            Multimedia, Edutainment, Games, Shells, SystemApps, Utils, Misc
        ]
        super(Root, self).__init__([x([]) for x in subs])

    def suitable(self, app):
        return True


def _menu(id, kws):
    class _Menu(Menu):
        name = id

        def suitable(self, app):
            return len(app.keywords & kws) > 0

    return _Menu


class KDevelop(Menu):
    name = "KDevelop"

    def suitable(self, app):
        return fnmatch.fnmatch(app.name, "*KDevelop*")


class DevApps(Menu):
    name = "Development"
    keywords = KwS(Development, Building, Debugger, IDE, GUIDesigner,
                   Profiling, RevisionControl, Translation, WebDevelopment)

    def __init__(self, subs):
        subs = [KDevelop]
        super(DevApps, self).__init__([x([]) for x in subs])


class Editors(Menu):
    name = "Editors"
    keywords = KwS(TextEditor)


class Misc(Menu):

    """Menu that contains the entries that didn't fit elsewhere"""
    name = "Other"

    def suitable(self, app):
        return True


class Shells(Menu):
    name = "Shells"
    keywords = KwS(Shell, TerminalEmulator, FileManager, Emulator)


class Utils(Menu):
    name = "Utilities"
    keywords = KwS(
        Utility, Accessibility, Calculator, Clock, Applet, Archiving, TrayIcon)

# KDE-specific keywords
#XKM    = Kw("X-KDE-More")
#XKI    = Kw("X-KDE-information")
#


class GNOMESettings(Menu):
    name = "GNOME"
    keywords = KwS(GNOME, Settings)

    def suitable(self, app):
        return len(app.keywords & self.keywords) == 2


def _kdeMenu(id, kws):
    class _Menu(Menu):
        name = id

        def suitable(self, app):
            return len(
                app.keywords & KwS(KDE, "X-KDE-settings")) and len(app.keywords & kws)

    return _Menu


class KDESettings(Menu):
    name = "KDE"

    def __init__(self, subs):
        subs = [
            ("Look & Feel", KwS("X-KDE-settings-looknfeel")),
            ("Sound", KwS("X-KDE-settings-sound")),
            ("System", KwS("X-KDE-settings-system")),
            ("Accessibility", KwS("X-KDE-settings-accessibility")),
            ("Components", KwS("X-KDE-settings-components")),
            ("Network", KwS("X-KDE-settings-network")),
            ("Security", KwS("X-KDE-settings-security")),
            ("Web", KwS("X-KDE-settings-webbrowsing")),
            ("Desktop", KwS("X-KDE-settings-desktop")),
            ("Peripherials", KwS("X-KDE-settings-peripherials")),
            ("Hardware", KwS("X-KDE-settings-hardware")),
            ("Power", KwS("X-KDE-settings-power"))
        ]
        super(KDESettings, self).__init__([_kdeMenu(*x)([]) for x in subs])

    def suitable(self, app):
        for x in app.keywords:
            if fnmatch.fnmatch(x, "X-KDE-settings*"):
                return True

        return False


class XfceSettings(Menu):
    name = "Xfce"

    def suitable(self, app):
        return len(app.keywords & KwS(Settings, Kw("X-XFCE"))) == 2


class SystemSettings(Menu):
    name = "Settings"
    keywords = KwS(Settings, DesktopSettings, HardwareSettings, Screensaver)

    def __init__(self, subs):
        subs = [GNOMESettings, KDESettings, XfceSettings]
        super(SystemSettings, self).__init__([x([]) for x in subs])


class SystemApps(Menu):
    name = "System"
    keywords = KwS(System, Filesystem, Monitor, Security, PackageManager)

    def __init__(self, subs):
        subs = [_menu("Information", KwS("X-KDE-information")), SystemSettings]
        super(SystemApps, self).__init__([x([]) for x in subs])


class ScienceApps(Menu):
    name = "Science"
    keywords = KwS(Electronics, Engineering, Science, Astronomy,
                   Biology, Chemistry, Geology, Math, MedicalSoftware, Physics)


class Edutainment(Menu):
    name = "Education"
    keywords = KwS(Education, Art, Construction, Music, Languages, Teaching)


class AdventureGames(Menu):
    name = "Adventure"
    keywords = KwS(AdventureGame, RolePlaying)


class ArcadeGames(Menu):
    name = "Arcade"
    keywords = KwS(ArcadeGame, ActionGame)


class BoardGames(Menu):
    name = "Board"
    keywords = KwS(BoardGame)


class CardGames(Menu):
    name = "Card"
    keywords = KwS(CardGame)


class PuzzleGames(Menu):
    name = "Puzzle"
    keywords = KwS(LogicGame, BlocksGame)


class SimGames(Menu):
    name = "Simulation"
    keywords = KwS(Simulation)


class SportsGames(Menu):
    name = "Sports"
    keywords = KwS(SportsGame)


class StrategyGames(Menu):
    name = "Strategy"
    keywords = KwS(StrategyGame)


class ToyGames(Menu):
    name = "Amusement"
    keywords = KwS(Amusement, KidsGame)


class Games(Menu):
    name = "Games"
    keywords = KwS(
        Amusement,
        Game,
        ActionGame,
        AdventureGame,
        ArcadeGame,
        BoardGame,
        BlocksGame,
        CardGame,
        KidsGame,
        LogicGame,
        RolePlaying,
        Simulation,
        SportsGame,
        StrategyGame)

    def __init__(self, subs):
        subs = [
            AdventureGames, ArcadeGames, BoardGames, CardGames,
            PuzzleGames, SimGames, SportsGames, StrategyGames, ToyGames
        ]
        super(Games, self).__init__([x([]) for x in subs])


class Multimedia(Menu):
    name = "Multimedia"
    keywords = KwS(AudioVideo, Audio, Midi, Mixer, Sequencer, Tuner,
                   Video, TV, AudioVideoEditing, Player, Recorder, DiscBurning)


class NetApps(Menu):
    name = "Network"
    keywords = KwS(
        Network,
        Dialup,
        InstantMessaging,
        IRCClient,
        FileTransfer,
        HamRadio,
        News,
        Email,
        P2P,
        RemoteAccess,
        Telephony,
        WebBrowser)


class GraphicApps(Menu):
    name = "Graphics"
    keywords = KwS(Graphics, x2DGraphics, x3DGraphics, VectorGraphics,
                   RasterGraphics, Scanning, OCR, Photograph, Viewer)


class OfficeApps(Menu):
    name = "Office"
    keywords = KwS(
        Office,
        Calendar,
        ContactManagement,
        Database,
        Dictionary,
        Chart,
        Finance,
        FlowChart,
        PDA,
        ProjectManagement,
        Presentation,
        Spreadsheet,
        WordProcessor)



import os.path
import Prophet

from Prophet.Legacy import App as _App, DropIn as _DropIn, ConsoleApp as _ConsoleApp, X11App as _X11App
from Keywords import Keyword as Kw, Set as KwS
from Prophet.Categories import *

from Keywords import Keyword as Kw, Set as KwS
from Prophet.Categories import *


class abook(_App, _ConsoleApp):
    name = "Abook"
    comment = "Address book"
    keywords = KwS(ContactManagement)


class alicq(_App, _X11App):
    name = "ALICQ"
    comment = "ICQ client for X"
    keywords = KwS(Network, InstantMessaging)


class althea(_App, _X11App):
    name = "Althea"
    comment = "E-mail client for X"
    keywords = KwS(Network, Email)


class aria(_App, _X11App):
    name = "Aria"
    comment = "Download manager"
    keywords = KwS(Network, FileTransfer, P2P)


class BitchX(_App, _ConsoleApp):  # FIXME : versioned
    name = "Bitch X"
    comment = "IRC client"
    keywords = KwS(Network, IRCClient)


class cadaver(_App, _ConsoleApp):  # FIXME ????
    name = "Cadaver"
    comment = "WebDAV client"
    keywords = KwS(Network)

# TODO : camstream, gqcam


class centericq(_App, _ConsoleApp):
    name = "CenterICQ"
    comment = "Console instant messenger"
    keywords = KwS(Network, InstantMessaging)


class dillo(_App, _X11App):
    name = "Dillo"
    comment = "Small & agile web browser"
    keywords = KwS(Network, WebBrowser, GTK)


class elinks(_App, _ConsoleApp):
    name = "ELinks"
    comment = "Text-mode web browser"
    keywords = KwS(Network, WebBrowser)


class elm(_App, _ConsoleApp):
    name = "Elm"
    comment = "Console mailer"
    keywords = KwS(Network, Email)


class everybuddy(_App, _X11App):
    name = "EveryBuddy"
    comment = "X11 instant messenger"
    keywords = KwS(Network, InstantMessaging)

# TODO : better detection


class firefox(_App, _X11App):
    name = "FireFox"
    comment = "Mozilla-derived web browser"
    keywords = KwS(Network, WebBrowser, GTK)
    exes = ["MozillaFirefox", "mozilla-firefox", "firefox"]


class ftp(_App, _ConsoleApp):
    name = "FTP"
    comment = "Console FTP client"
    keywords = KwS(Network, FileTransfer, Core)


class icqnix(_App, _X11App):
    name = "ICQnix"
    comment = "ICQ client for X"
    keywords = KwS(Network, InstantMessaging)


class irssi(_App, _ConsoleApp):
    name = "Irssi"
    comment = "IRC client"
    keywords = KwS(Network, IRCClient)


class licq(_App, _X11App):
    name = "Licq"
    comment = "ICQ client for X"
    keywords = KwS(Network, InstantMessaging)


class links(_App, _ConsoleApp):
    name = "Links"
    comment = "Text-mode web browser"
    keywords = KwS(Network, WebBrowser)


class LinNeighbohrhood(_App, _X11App):
    name = "Lin... SAMBA browser"
    comment = "LinNxxx SAMBA browser"
    keywords = KwS(Network, FileTransfer, RemoteAccess)


class linpopup(_App, _X11App):
    name = "LinPopUp"
    comment = "Program to communicate with Windows"
    keywords = KwS(Network, RemoteAccess)
    exes = ["LinPopUp", "linpopup"]


class lftp(_App, _ConsoleApp):
    name = "LFTP"
    comment = "Console FTP client"
    keywords = KwS(Network, FileTransfer)


class lynx(_App, _ConsoleApp):
    name = "Lynx"
    comment = "Text-mode web browser"
    keywords = KwS(Network, WebBrowser)


class mail(_App, _ConsoleApp):
    name = "Mail"
    comment = "Console mailer"
    keywords = KwS(Network, Email, Core)

# TODO : better detection


class mozilla(_App, _X11App):
    name = "Mozilla"
    comment = "Mozilla web suite"
    keywords = KwS(Network, WebBrowser, Email, WebDevelopment, GTK)


class mutt(_App, _ConsoleApp):
    name = "Mutt"
    comment = "Mail user agent"
    keywords = KwS(Network, Email)


class ncftp(_App, _ConsoleApp):
    name = "NcFTP"
    comment = "Improved FTP client"
    keywords = KwS(Network, FileTransfer)


class netscape(_App, _X11App):
    name = "Netscape"
    comment = "Netscape web suite"
    keywords = KwS(Network, WebBrowser, Email, WebDevelopment)


class opera(_DropIn, _X11App):
    name = "Opera"
    comment = "Opera browser"
    keywords = KwS(WebBrowser, Email)

    def relevant(self, path):
        if Prophet.isExe(os.path.join(path, "opera")):
            raise opera.StopDescention(path)


class paddress(_App, _X11App):
    name = "Paddress"
    comment = "Address book"
    keywords = KwS(ContactManagement)


class pan(_App, _X11App):
    name = "Pan"
    comment = "News agent"
    keywords = KwS(Network, News)


class pine(_App, _ConsoleApp):
    name = "Pine"
    comment = "Mail/news agent"
    keywords = KwS(Network, Email, News)


class postilion(_App, _X11App):
    name = "Postilion"
    comment = "Tcl/Tk mail client"
    keywords = KwS(Network, Email)


class skipstone(_App, _X11App):
    name = "Skipstone"
    comment = "Mozilla-derived web browser"
    keywords = KwS(Network, WebBrowser)


class spruce(_App, _X11App):
    name = "Spruce"
    comment = "X11 mail client"
    keywords = KwS(Network, Email)


class slrn(_App, _X11App):
    name = "SLRN"
    comment = "USENET newreader"
    keywords = KwS(Network, News)


class talk(_App, _ConsoleApp):  # FIXME : is it text-mode?
    name = "Talk"
    comment = "Talk communication client"
    keywords = KwS(Network, IRCClient)


class telnet(_App, _ConsoleApp):
    name = "Telnet"
    comment = "Standard Telnet client"
    keywords = KwS(Network, Shell, Core)


class thunderbird(_App, _X11App):
    name = "Thunderbird"
    comment = "Mozilla news/mail client"
    keywords = KwS(Network, News, Email)
    exes = ["MozillaThunderbird", "thunderbird"]


class tkpppoe(_App, _X11App):
    name = "TkPPPoE"
    comment = "PPP frontend"
    keywords = KwS(Network, Dialup)


class tkrat(_App, _X11App):
    name = "TkRat"
    comment = "Tcl/Tk mail client"
    keywords = KwS(Network, Email)


class w3m(_App, _ConsoleApp):
    name = "W3m"
    comment = "Text-mode web browser"
    keywords = KwS(Network, WebBrowser)


class xarchie(_App, _X11App):
    name = "Xarchie"
    comment = "Archie client"
    keywords = KwS(Network)


class xchat(_App, _X11App):
    name = "X-Chat"
    comment = "Powerful chat client"
    keywords = KwS(Network, InstantMessaging)


class xmailbox(_App, _X11App):
    name = "Xmailbox"
    comment = "Mail delivery notifier"
    keywords = KwS(Network, Email)


class ytalk(_App, _ConsoleApp):
    name = "Ytalk"
    comment = "Talk compatible communication client"
    keywords = KwS(Network, IRCClient)
